from .guild_join_request_form_response import *

__all__ = guild_join_request_form_response.__all__

from ....utils.module_deprecation import deprecated_import

deprecated_import(GuildJoinRequestFormResponse, 'GuildRequestFormResponse')
