__all__ = ()

LIMIT_GLOBAL = 100
LIMIT_GUILD = 100
LENGTH_MAX = 4000

NAME_LENGTH_MIN = 1
NAME_LENGTH_MAX = 32
DESCRIPTION_LENGTH_MIN = 2
DESCRIPTION_LENGTH_MAX = 100
OPTIONS_MAX = 25
CHOICES_MAX = 25
